@@ -47,6 +47,12 @@
 #include <systypes.h>
 #endif
 
+typedef enum prim_num {
+/* definitions of N_execute etc. */
+#include "prim_num.i"
+  N_START_SUPER
+} PrimNum;
+
 /* global variables for engine.c 
    We put them here because engine.c is compiled several times in
    different ways for the same engine. */
@@ -54,6 +60,21 @@ Cell *SP;
 Float *FP;
 Address UP=NULL;
 
+#ifdef HAS_FFCALL
+Cell *RP;
+Address LP;
+
+#include <callback.h>
+
+va_alist clist;
+
+void engine_callback(Xt* fcall, void * alist)
+{
+  clist = (va_alist)alist;
+  engine(fcall, SP, RP, FP, LP);
+}
+#endif
+
 #ifdef GFORTH_DEBUGGING
 /* define some VM registers as global variables, so they survive exceptions;
    global register variables are not up to the task (according to the 
@@ -131,6 +152,9 @@ Cell last_jump=0; /* if the last prim was compiled without jump, this
 static int no_super=0;   /* true if compile_prim should not fuse prims */
 static int no_dynamic=NO_DYNAMIC_DEFAULT; /* if true, no code is generated
 					     dynamically */
+static int print_metrics=0; /* if true, print metrics on exit */
+static int static_super_number = 10000000; /* number of ss used if available */
+static int ss_greedy = 0; /* if true: use greedy, not optimal ss selection */
 
 #ifdef HAS_DEBUG
 int debug=0;
@@ -183,16 +207,21 @@ int gforth_memcmp(const char * s1, const char * s2, size_t n)
  * bits 8..0 of a primitive token index into the group
  */
 
-static Cell groups[32] = {
+Cell groups[32] = {
   0,
+  0
 #undef GROUP
-#define GROUP(x, n) DOESJUMP+1+n,
+#undef GROUPADD
+#define GROUPADD(n) +n
+#define GROUP(x, n) , 0
 #include "prim_grp.i"
 #undef GROUP
+#undef GROUPADD
 #define GROUP(x, n)
+#define GROUPADD(n)
 };
 
-void relocate(Cell *image, const char *bitstring, 
+void relocate(Cell *image, const unsigned char *bitstring, 
               int size, Cell base, Label symbols[])
 {
   int i=0, j, k, steps=(size/sizeof(Cell))/RELINFOBITS;
@@ -206,10 +235,19 @@ void relocate(Cell *image, const char *bitstring,
   Cell *start = (Cell * ) (((void *) image) - ((void *) base));
 
   /* group index into table */
+  if(groups[31]==0) {
+    int groupsum=0;
+    for(i=0; i<32; i++) {
+      groupsum += groups[i];
+      groups[i] = groupsum;
+      /* printf("group[%d]=%d\n",i,groupsum); */
+    }
+    i=0;
+  }
   
 /* printf("relocating to %x[%x] start=%x base=%x\n", image, size, start, base); */
   
-  for (max_symbols=DOESJUMP+1; symbols[max_symbols]!=0; max_symbols++)
+  for (max_symbols=0; symbols[max_symbols]!=0; max_symbols++)
     ;
   max_symbols--;
   size/=sizeof(Cell);
@@ -247,7 +285,7 @@ void relocate(Cell *image, const char *bitstring,
 		  compile_prim1(&image[i]);
 #endif
 	      } else
-		fprintf(stderr,"Primitive %ld used in this image at $%lx is not implemented by this\n engine (%s); executing this code will crash.\n",(long)CF(token),(long)&image[i],PACKAGE_VERSION);
+		fprintf(stderr,"Primitive %ld used in this image at $%lx (offset $%x) is not implemented by this\n engine (%s); executing this code will crash.\n",(long)CF(token),(long)&image[i], i, PACKAGE_VERSION);
 	    }
 	  } else {
 	    int tok = -token & 0x1FF;
@@ -262,7 +300,7 @@ void relocate(Cell *image, const char *bitstring,
 		compile_prim1(&image[i]);
 #endif
 	    } else
-	      fprintf(stderr,"Primitive %lx, %d of group %d used in this image at $%lx is not implemented by this\n engine (%s); executing this code will crash.\n", (long)-token, tok, group, (long)&image[i],PACKAGE_VERSION);
+	      fprintf(stderr,"Primitive %lx, %d of group %d used in this image at $%lx (offset $%x) is not implemented by this\n engine (%s); executing this code will crash.\n", (long)-token, tok, group, (long)&image[i],i,PACKAGE_VERSION);
 	  }
 	} else {
           /* if base is > 0: 0 is a null reference so don't adjust*/
@@ -380,7 +418,7 @@ Address my_alloc(Cell size)
   Address r;
 
   r=alloc_mmap(size);
-  if (r!=MAP_FAILED)
+  if (r!=(Address)MAP_FAILED)
     return r;
 #endif /* HAVE_MMAP */
   /* use malloc as fallback */
@@ -400,7 +438,7 @@ Address dict_alloc_read(FILE *file, Cell imagesize, Cell dictsize, Cell offset)
     after_alloc(image,dictsize);
   }
 #endif /* defined(HAVE_MMAP) */
-  if (image == MAP_FAILED) {
+  if (image == (Address)MAP_FAILED) {
     image = my_alloc(dictsize+offset)+offset;
     rewind(file);  /* fseek(imagefile,0L,SEEK_SET); */
     fread(image, 1, imagesize, file);
@@ -515,6 +553,86 @@ void print_sizes(Cell sizebyte)
 	  1 << ((sizebyte >> 5) & 3));
 }
 
+/* static superinstruction stuff */
+
+struct cost {
+  char loads;       /* number of stack loads */
+  char stores;      /* number of stack stores */
+  char updates;     /* number of stack pointer updates */
+  short offset;      /* offset into super2 table */
+  char length;      /* number of components */
+};
+
+PrimNum super2[] = {
+#include "super2.i"
+};
+
+struct cost super_costs[] = {
+#include "costs.i"
+};
+
+#define HASH_SIZE 256
+
+struct super_table_entry {
+  struct super_table_entry *next;
+  PrimNum *start;
+  short length;
+  PrimNum super;
+} *super_table[HASH_SIZE];
+int max_super=2;
+
+int hash_super(PrimNum *start, int length)
+{
+  int i, r;
+  
+  for (i=0, r=0; i<length; i++) {
+    r <<= 1;
+    r += start[i];
+  }
+  return r & (HASH_SIZE-1);
+}
+
+int lookup_super(PrimNum *start, int length)
+{
+  int hash=hash_super(start,length);
+  struct super_table_entry *p = super_table[hash];
+
+  assert(length >= 2);
+  for (; p!=NULL; p = p->next) {
+    if (length == p->length &&
+	memcmp((char *)p->start, (char *)start, length*sizeof(PrimNum))==0)
+      return p->super;
+  }
+  return -1;
+}
+
+void prepare_super_table()
+{
+  int i;
+  int nsupers = 0;
+
+  for (i=0; i<sizeof(super_costs)/sizeof(super_costs[0]); i++) {
+    struct cost *c = &super_costs[i];
+    if (c->length > 1 && nsupers < static_super_number) {
+      int hash = hash_super(super2+c->offset, c->length);
+      struct super_table_entry **p = &super_table[hash];
+      struct super_table_entry *e = malloc(sizeof(struct super_table_entry));
+      e->next = *p;
+      e->start = super2 + c->offset;
+      e->length = c->length;
+      e->super = i;
+      *p = e;
+      if (c->length > max_super)
+	max_super = c->length;
+      nsupers++;
+    }
+  }
+  if (debug)
+    fprintf(stderr, "Using %d static superinsts\n", nsupers);
+}
+
+/* dynamic replication/superinstruction stuff */
+
 #define MAX_IMMARGS 2
 
 #ifndef NO_DYNAMIC
@@ -546,19 +664,46 @@ int compare_priminfo_length(const void *_a, const void *_b)
           of (char) and @ instead of >code-address */
     return (*b)->start - (*a)->start;
 }
+#endif /* !defined(NO_DYNAMIC) */
+
+static char superend[]={
+#include "prim_superend.i"
+};
 
-#endif /* defined(NO_DYNAMIC) */
 Cell npriminfos=0;
 
+int compare_labels(const void *pa, const void *pb)
+{
+  Label a = *(Label *)pa;
+  Label b = *(Label *)pb;
+  return a-b;
+}
+
+Label bsearch_next(Label key, Label *a, UCell n)
+     /* a is sorted; return the label >=key that is the closest in a;
+        return NULL if there is no label in a >=key */
+{
+  int mid = (n-1)/2;
+  if (n<1)
+    return NULL;
+  if (n == 1) {
+    if (a[0] < key)
+      return NULL;
+    else
+      return a[0];
+  }
+  if (a[mid] < key)
+    return bsearch_next(key, a+mid+1, n-mid-1);
+  else
+    return bsearch_next(key, a, mid+1);
+}
 
 void check_prims(Label symbols1[])
 {
   int i;
 #ifndef NO_DYNAMIC
-  Label *symbols2, *symbols3, *ends1;
-  static char superend[]={
-#include "prim_superend.i"
-  };
+  Label *symbols2, *symbols3, *ends1, *ends1j, *ends1jsorted;
+  int nends1j;
 #endif
 
   if (debug)
@@ -569,7 +714,7 @@ void check_prims(Label symbols1[])
 #define str(s) #s
   fprintf(stderr, "Compiled with gcc-" xstr(__GNUC__) "." xstr(__GNUC_MINOR__) "\n"); 
 #endif
-  for (i=DOESJUMP+1; symbols1[i+1]!=0; i++)
+  for (i=0; symbols1[i]!=0; i++)
     ;
   npriminfos = i;
   
@@ -582,28 +727,57 @@ void check_prims(Label symbols1[])
 #else
   symbols3=symbols1;
 #endif
-  ends1 = symbols1+i+1-DOESJUMP;
+  ends1 = symbols1+i+1;
+  ends1j =   ends1+i;
+  nends1j = i+1;
+  ends1jsorted = (Label *)alloca(nends1j*sizeof(Label));
+  memcpy(ends1jsorted,ends1j,nends1j*sizeof(Label));
+  qsort(ends1jsorted, nends1j, sizeof(Label), compare_labels);
+  
   priminfos = calloc(i,sizeof(PrimInfo));
-  for (i=DOESJUMP+1; symbols1[i+1]!=0; i++) {
+  for (i=0; symbols1[i]!=0; i++) {
     int prim_len = ends1[i]-symbols1[i];
     PrimInfo *pi=&priminfos[i];
     int j=0;
     char *s1 = (char *)symbols1[i];
     char *s2 = (char *)symbols2[i];
     char *s3 = (char *)symbols3[i];
+    Label endlabel = bsearch_next(symbols1[i]+1,ends1jsorted,nends1j);
 
     pi->start = s1;
-    pi->superend = superend[i-DOESJUMP-1]|no_super;
+    pi->superend = superend[i]|no_super;
     if (pi->superend)
-      pi->length = symbols1[i+1]-symbols1[i];
+      pi->length = endlabel-symbols1[i];
     else
       pi->length = prim_len;
-    pi->restlength = symbols1[i+1] - symbols1[i] - pi->length;
+    pi->restlength = endlabel - symbols1[i] - pi->length;
     pi->nimmargs = 0;
     if (debug)
       fprintf(stderr, "Prim %3d @ %p %p %p, length=%3ld restlength=%2ld superend=%1d",
 	      i, s1, s2, s3, (long)(pi->length), (long)(pi->restlength), pi->superend);
+    if (endlabel == NULL) {
+      pi->start = NULL; /* not relocatable */
+      if (pi->length<0) pi->length=100;
+      if (debug)
+	fprintf(stderr,"\n   non_reloc: no J label > start found\n");
+      continue;
+    }
+    if (ends1[i] > endlabel && !pi->superend) {
+      pi->start = NULL; /* not relocatable */
+      pi->length = endlabel-symbols1[i];
+      if (debug)
+	fprintf(stderr,"\n   non_reloc: there is a J label before the K label (restlength<0)\n");
+      continue;
+    }
+    if (ends1[i] < pi->start && !pi->superend) {
+      pi->start = NULL; /* not relocatable */
+      pi->length = endlabel-symbols1[i];
+      if (debug)
+	fprintf(stderr,"\n   non_reloc: K label before I label (length<0)\n");
+      continue;
+    }
     assert(prim_len>=0);
+    assert(pi->restlength >=0);
     while (j<(pi->length+pi->restlength)) {
       if (s1[j]==s3[j]) {
 	if (s1[j] != s2[j]) {
@@ -732,6 +906,21 @@ int forget_dyncode(Address code)
 #endif /* !defined(NO_DYNAMIC) */
 }
 
+long dyncodesize(void)
+{
+#ifndef NO_DYNAMIC
+  struct code_block_list *p;
+  long size=0;
+  for (p=code_block_list; p!=NULL; p=p->next) {
+    if (code_here >= p->block && code_here < p->block+p->size)
+      return size + (code_here - p->block);
+    else
+      size += p->size;
+  }
+#endif /* !defined(NO_DYNAMIC) */
+  return 0;
+}
+
 Label decompile_code(Label _code)
 {
 #ifdef NO_DYNAMIC
@@ -752,7 +941,8 @@ Label decompile_code(Label _code)
   for (i=npriminfos-1; i>DOESJUMP; i--) {
     PrimInfo *pi=decomp_prims[i];
     if (pi->start==code || (pi->start && memcmp(code,pi->start,pi->length)==0))
-      return pi->start;
+      return vm_prims[super2[super_costs[pi-priminfos].offset]];
+    /* return pi->start;*/
   }
   return code;
 #endif /* !defined(NO_DYNAMIC) */
@@ -772,9 +962,6 @@ struct doesexecinfo {
   Cell *xt; /* cfa of word whose does-code needs calling */
 } doesexecinfos[10000];
 
-/* definitions of N_execute etc. */
-#include "prim_num.i"
-
 void set_rel_target(Cell *source, Label target)
 {
   *source = ((Cell)target)-(((Cell)source)+4);
@@ -830,19 +1017,11 @@ void finish_code(void)
   flush_to_here();
 }
 
-void compile_prim1(Cell *start)
+#if 0
+/* compile *start into a dynamic superinstruction, updating *start */
+void compile_prim_dyn(Cell *start)
 {
-#if defined(DOUBLY_INDIRECT)
-  Label prim=(Label)*start;
-  if (prim<((Label)(xts+DOESJUMP)) || prim>((Label)(xts+npriminfos))) {
-    fprintf(stderr,"compile_prim encountered xt %p\n", prim);
-    *start=(Cell)prim;
-    return;
-  } else {
-    *start = (Cell)(prim-((Label)xts)+((Label)vm_prims));
-    return;
-  }
-#elif defined(NO_IP)
+#if defined(NO_IP)
   static Cell *last_start=NULL;
   static Xt last_prim=NULL;
   /* delay work by one call in order to get relocated immargs */
@@ -927,22 +1106,168 @@ void compile_prim1(Cell *start)
   return;
 #endif /* !defined(DOUBLY_INDIRECT) */
 }
+#endif /* 0 */
 
-Label compile_prim(Label prim)
+Cell compile_prim_dyn(unsigned p)
 {
-  Cell x=(Cell)prim;
-  assert(0);
-  compile_prim1(&x);
-  return (Label)x;
+  Cell static_prim = (Cell)vm_prims[p];
+#if defined(NO_DYNAMIC)
+  return static_prim;
+#else /* !defined(NO_DYNAMIC) */
+  Address old_code_here;
+
+  if (no_dynamic)
+    return static_prim;
+  if (p>=npriminfos || priminfos[p].start == 0) { /* not a relocatable prim */
+    append_jump();
+    return static_prim;
+  }
+  old_code_here = append_prim(p);
+  last_jump = (priminfos[p].superend) ? 0 : p;
+  return (Cell)old_code_here;
+#endif  /* !defined(NO_DYNAMIC) */
 }
 
-#if defined(PRINT_SUPER_LENGTHS) && !defined(NO_DYNAMIC)
-Cell prim_length(Cell prim)
+#ifndef NO_DYNAMIC
+int cost_codesize(int prim)
 {
-  return priminfos[prim+DOESJUMP+1].length;
+  return priminfos[prim].length;
 }
 #endif
 
+int cost_ls(int prim)
+{
+  struct cost *c = super_costs+prim;
+
+  return c->loads + c->stores;
+}
+
+int cost_lsu(int prim)
+{
+  struct cost *c = super_costs+prim;
+
+  return c->loads + c->stores + c->updates;
+}
+
+int cost_nexts(int prim)
+{
+  return 1;
+}
+
+typedef int Costfunc(int);
+Costfunc *ss_cost =  /* cost function for optimize_bb */
+#ifdef NO_DYNAMIC
+cost_lsu;
+#else
+cost_codesize;
+#endif
+
+struct {
+  Costfunc *costfunc;
+  char *metricname;
+  long sum;
+} cost_sums[] = {
+#ifndef NO_DYNAMIC
+  { cost_codesize, "codesize", 0 },
+#endif
+  { cost_ls,       "ls",       0 },
+  { cost_lsu,      "lsu",      0 },
+  { cost_nexts,    "nexts",    0 }
+};
+
+#define MAX_BB 128 /* maximum number of instructions in BB */
+
+/* use dynamic programming to find the shortest paths within the basic
+   block origs[0..ninsts-1]; optimals[i] contains the superinstruction
+   on the shortest path to the end of the BB */
+void optimize_bb(PrimNum origs[], PrimNum optimals[], int ninsts)
+{
+  int i,j, mincost;
+  static int costs[MAX_BB+1];
+
+  assert(ninsts<MAX_BB);
+  costs[ninsts]=0;
+  for (i=ninsts-1; i>=0; i--) {
+    optimals[i] = origs[i];
+    costs[i] = mincost = costs[i+1] + ss_cost(optimals[i]);
+    for (j=2; j<=max_super && i+j<=ninsts ; j++) {
+      int super, jcost;
+
+      super = lookup_super(origs+i,j);
+      if (super >= 0) {
+	jcost = costs[i+j] + ss_cost(super);
+	if (jcost <= mincost) {
+	  optimals[i] = super;
+	  mincost = jcost;
+	  if (!ss_greedy)
+	    costs[i] = jcost;
+	}
+      }
+    }
+  }
+}
+
+/* rewrite the instructions pointed to by instps to use the
+   superinstructions in optimals */
+void rewrite_bb(Cell *instps[], PrimNum *optimals, int ninsts)
+{
+  int i,j, nextdyn;
+  Cell inst;
+
+  for (i=0, nextdyn=0; i<ninsts; i++) {
+    if (i==nextdyn) { /* compile dynamically */
+      nextdyn += super_costs[optimals[i]].length;
+      inst = compile_prim_dyn(optimals[i]);
+      for (j=0; j<sizeof(cost_sums)/sizeof(cost_sums[0]); j++)
+	cost_sums[j].sum += cost_sums[j].costfunc(optimals[i]);
+    } else { /* compile statically */
+      inst = (Cell)vm_prims[optimals[i]];
+    }
+    *(instps[i]) = inst;
+  }
+}
+
+/* compile *start, possibly rewriting it into a static and/or dynamic
+   superinstruction */
+void compile_prim1(Cell *start)
+{
+#if defined(DOUBLY_INDIRECT)
+  Label prim=(Label)*start;
+  if (prim<((Label)(xts+DOESJUMP)) || prim>((Label)(xts+npriminfos))) {
+    fprintf(stderr,"compile_prim encountered xt %p\n", prim);
+    *start=(Cell)prim;
+    return;
+  } else {
+    *start = (Cell)(prim-((Label)xts)+((Label)vm_prims));
+    return;
+  }
+#elif defined(INDIRECT_THREADED)
+  return;
+#else /* !(defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED)) */
+  static Cell *instps[MAX_BB];
+  static PrimNum origs[MAX_BB];
+  static PrimNum optimals[MAX_BB];
+  static int ninsts=0;
+  PrimNum prim_num;
+
+  if (start==NULL)
+    goto end_bb;
+  prim_num = ((Xt)*start)-vm_prims;
+  if (prim_num >= npriminfos)
+    goto end_bb;
+  assert(ninsts<MAX_BB);
+  instps[ninsts] = start;
+  origs[ninsts] = prim_num;
+  ninsts++;
+  if (ninsts >= MAX_BB || superend[prim_num]) {
+  end_bb:
+    optimize_bb(origs,optimals,ninsts);
+    rewrite_bb(instps,optimals,ninsts);
+    ninsts=0;
+  }
+#endif /* !(defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED)) */
+}
+
 Address loader(FILE *imagefile, char* filename)
 /* returns the address of the image proper (after the preamble) */
 {
@@ -973,6 +1298,7 @@ Address loader(FILE *imagefile, char* filename)
 
   vm_prims = engine(0,0,0,0,0);
   check_prims(vm_prims);
+  prepare_super_table();
 #ifndef DOUBLY_INDIRECT
 #ifdef PRINT_SUPER_LENGTHS
   print_super_lengths();
@@ -1160,6 +1486,14 @@ UCell convsize(char *s, UCell elemsize)
   return n*m;
 }
 
+enum {
+  ss_number = 256,
+  ss_min_codesize,
+  ss_min_ls,
+  ss_min_lsu,
+  ss_min_nexts,
+};
+
 void gforth_args(int argc, char ** argv, char ** path, char ** imagename)
 {
   int c;
@@ -1187,6 +1521,15 @@ void gforth_args(int argc, char ** argv, char ** path, char ** imagename)
       {"no-super", no_argument, &no_super, 1},
       {"no-dynamic", no_argument, &no_dynamic, 1},
       {"dynamic", no_argument, &no_dynamic, 0},
+      {"print-metrics", no_argument, &print_metrics, 1},
+      {"ss-number", required_argument, NULL, ss_number},
+#ifndef NO_DYNAMIC
+      {"ss-min-codesize", no_argument, NULL, ss_min_codesize},
+#endif
+      {"ss-min-ls",       no_argument, NULL, ss_min_ls},
+      {"ss-min-lsu",      no_argument, NULL, ss_min_lsu},
+      {"ss-min-nexts",    no_argument, NULL, ss_min_nexts},
+      {"ss-greedy",       no_argument, &ss_greedy, 1},
       {0,0,0,0}
       /* no-init-file, no-rc? */
     };
@@ -1210,6 +1553,13 @@ void gforth_args(int argc, char ** argv, char ** path, char ** imagename)
     case 's': die_on_signal = 1; break;
     case 'x': debug = 1; break;
     case 'v': fputs(PACKAGE_STRING"\n", stderr); exit(0);
+    case ss_number: static_super_number = atoi(optarg); break;
+#ifndef NO_DYNAMIC
+    case ss_min_codesize: ss_cost = cost_codesize; break;
+#endif
+    case ss_min_ls:       ss_cost = cost_ls;       break;
+    case ss_min_lsu:      ss_cost = cost_lsu;      break;
+    case ss_min_nexts:    ss_cost = cost_nexts;    break;
     case 'h': 
       fprintf(stderr, "Usage: %s [engine options] ['--'] [image arguments]\n\
 Engine Options:\n\
@@ -1229,7 +1579,14 @@ Engine Options:\n\
   --no-super                        No dynamically formed superinstructions\n\
   --offset-image		    Load image at a different position\n\
   -p PATH, --path=PATH		    Search path for finding image and sources\n\
+  --print-metrics		    Print some code generation metrics on exit\n\
   -r SIZE, --return-stack-size=SIZE Specify return stack size\n\
+  --ss-greedy                       greedy, not optimal superinst selection\n\
+  --ss-min-codesize                 select superinsts for smallest native code\n\
+  --ss-min-ls                       minimize loads and stores\n\
+  --ss-min-lsu                      minimize loads, stores, and pointer updates\n\
+  --ss-min-nexts                    minimize the number of static superinsts\n\
+  --ss-number=N                     use N static superinsts (default max)\n\
   -v, --version			    Print engine version and exit\n\
 SIZE arguments consist of an integer followed by a unit. The unit can be\n\
   `b' (byte), `e' (element; default), `k' (KB), `M' (MB), `G' (GB) or `T' (TB).\n",
@@ -1282,7 +1639,15 @@ int main(int argc, char **argv, char **env)
 
 #ifdef HAS_OS
   gforth_args(argc, argv, &path, &imagename);
-#endif
+#ifndef NO_DYNAMIC
+  if (no_dynamic && ss_cost == cost_codesize) {
+    ss_cost = cost_nexts;
+    cost_sums[0] = cost_sums[1]; /* don't use cost_codesize for print-metrics */
+    if (debug)
+      fprintf(stderr, "--no-dynamic conflicts with --ss-min-codesize, reverting to --ss-min-nexts\n");
+  }
+#endif /* !defined(NO_DYNAMIC) */
+#endif /* defined(HAS_OS) */
 
 #ifdef INCLUDE_IMAGE
   set_stack_sizes((ImageHeader *)image);
@@ -1325,5 +1690,12 @@ int main(int argc, char **argv, char **env)
 #endif
     deprep_terminal();
   }
+  if (print_metrics) {
+    int i;
+    fprintf(stderr, "code size = %8ld\n", dyncodesize());
+    for (i=0; i<sizeof(cost_sums)/sizeof(cost_sums[0]); i++)
+      fprintf(stderr, "metric %8s: %8ld\n",
+	      cost_sums[i].metricname, cost_sums[i].sum);
+  }
   return retvalue;
 }
